#include <LEDA/ugraph.h>
#include <Tim/graph.h>
#include "lbfs.h"
#include "interval.h"

char lbfs_version[] = "$Id: lbfs.cc,v 1.5 1996/08/22 21:24:36 tbm Exp $";

/*****************************************************************************/

bool operator == (const vertex_order &order1, const vertex_order &order2) {
	node v;
	forall (v, order1.nodes) {
		if (order1.position[v] != order2.position[v]) {
			return false;
		}
	}
	return true;
}

/*****************************************************************************/

ostream & operator << (ostream &out, const vertex_order &order) {
	if (order.the_graph == NULL) {
		return out;
	}

	node v;
	bool first = true;
	forall (v, order.nodes) {
		if (first) {
			first = false;
		} else {
			out << " ";
		}
		out << (*order.the_graph)[v].label;
	}
	return out;
}

/*****************************************************************************/

ostream & operator << (ostream &out, const eq_class &set) {
	if (set.the_graph == NULL) {
		return out;
	}

	node v;
	bool first = true;
	out << "{";
	forall (v, set.S) {
		if (first) {
			first = false;
		} else {
			out << " ";
		}
		out << (*set.the_graph)[v].label;
	}
	return out << "}";
}

/*****************************************************************************/

ostream & operator << (ostream &out, const lexicographic_order &order) {
	if (order.the_graph == NULL) {
		return out;
	}

	eq_class *set;
	bool first = true;
	forall (set, order.order) {
		if (first) {
			first = false;
		} else {
			out << " ";
		}
		out << *set;
	}
	return out;
}

/*****************************************************************************/
/* Lexicographic Order                                                       */
/*****************************************************************************/

void lexicographic_order::init (const labelled_graph &G) {
	the_graph = &G;
	eq_classes.init (G);
	eq_class_entry.init (G);
	visited.init (G, false);
	refinements = 0;

	eq_class *order_node = new eq_class (G, refinements);
	list_item the_eq_class;

	if ((G.number_of_nodes () > 0) && (order_node != NULL)) {
		order_node->creation_timestamp = refinements;

		the_eq_class = order.append (order_node);

		node v;
		forall_nodes (v, G) {
			eq_classes[v]     = the_eq_class;
			eq_class_entry[v] = order_node->S.append (v);
		}
	}
}

void lexicographic_order::refine (const list<node> L) {
	node v;
	list_item new_eq_class_item;
	eq_class *new_eq_class;
	list_item old_eq_class_item;
	eq_class *old_eq_class;

	refinements++;

	forall (v, L) {
		if (!visited[v]) {
			old_eq_class_item = eq_classes[v];
			old_eq_class = order.contents (old_eq_class_item);

			new_eq_class_item = order.pred (old_eq_class_item);
			if (new_eq_class_item != NULL) {
				new_eq_class = order.contents (new_eq_class_item);
			}

			if ((new_eq_class_item == NULL) ||
			    (new_eq_class->creation_timestamp != refinements)) {
				if (old_eq_class->S.size () == 1) {
					// Don't bother moving set of size 1
					old_eq_class->creation_timestamp--;
					continue;
				}
				new_eq_class = new eq_class (*the_graph, refinements);
				new_eq_class_item = order.insert (new_eq_class, old_eq_class_item, before);
			}

			old_eq_class->S.del_item (eq_class_entry[v]);

			eq_classes[v] = new_eq_class_item;
			eq_class_entry[v] = new_eq_class->S.append (v);

			if (old_eq_class->S.empty ()) {
				delete order.del_item (old_eq_class_item);
				// Ensure nothing more gets put in new eq class
				new_eq_class->creation_timestamp--;
			}
		}
	}
}

void lexicographic_order::extract (node v) {
	if (!visited[v]) {
		eq_class *order_node = order.contents (eq_classes[v]);
		order_node->S.del_item (eq_class_entry[v]);
		if (order_node->S.empty ()) {
			delete order.del_item (eq_classes[v]);
		}
		visited[v] = true;
	}
}

node lexicographic_order::extract (tiebreaker tb, void *tbinfo1, void *tbinfo2) {
	node v = NULL;
	if (!order.empty ()) {
		v = tb (*the_graph, *this, order.head ()->S, tbinfo1, tbinfo2);
		if (v != NULL) {
			extract (v);
		}
	}
	return v;
}

/*****************************************************************************/
/* Tiebreaker that chooses arbitrarily                                       */
/*****************************************************************************/

node arbitrary (
	const labelled_graph &G,
	const lexicographic_order &order,
	const list<node> &choices,
	void *info1,
	void *info2
) {
	return choices.head ();
}

/*****************************************************************************/
/* Lexicographic Breadth-First Search                                        */
/*****************************************************************************/

node LBFS (
	const labelled_graph &G,
	node v,
	vertex_order &order,
	tiebreaker tb,
	void *tbinfo1,
	void *tbinfo2
) {
	if (G.number_of_nodes () == 0) {
		return NULL;
	}

	lexicographic_order lex_order (G);

	lex_order.extract (v);

	if (options.full_info) {
		cout << "\n\t\tVertex\tUnvisited vertices\n";
	}

	for (int i = G.number_of_nodes ();; i--) {
		order.position[v] = G.number_of_nodes () + 1 - i;
		order.position_item[v] = order.nodes.append (v);
		lex_order.refine (G.adj_nodes (v));
		if (options.full_info) {
			cout << "\t\t" << G[v].label << "\t" << lex_order << "\n";
		}
		if (i == 1) {
			break;
		}
		v = lex_order.extract (tb, tbinfo1, tbinfo2);
	}
	return v;
}
