#include <stdio.h>
#include <minmax.h>
#include <LEDA/p_queue.h>
#include <LEDA/graph_edit.h>
#include <Tim/graph.h>

// For the parser
extern int yyparse(void);

// Graph and attributes to load
GAGraph   *parsed_graph;
KeyMapper keymapper;

/*****************************************************************************/

#if 0
void Read (labelled_node &info, istream& in) {
}
#endif

/*****************************************************************************/

#if 0
void Print (const labelled_node &info, ostream &out) {
	out << info.p << " " << info.label << "\n";
}
#endif

/*****************************************************************************/

void labelled_graph::edit (string filename) {
#if 0
	window W (500, 500);
	graph_edit (W, *this, false);
#else
	graph_edit (*this, false, filename);
#endif
}

/*****************************************************************************/

int labelled_graph::read_GraphEd (char *filename) {
	extern FILE *yyin;
	if ((yyin = fopen (filename, "r")) == NULL) {
		return 1;
	}

	parsed_graph = &GAG;
	yyparse ();
	fclose (yyin);
	yyin = NULL;

	int i = 0;
	double x, y;
	double maxx = 0, maxy = 0, miny = MAXINT, minx = MAXINT;
	node v;
	GAList *attrs;
	GA *attr;
	KeyType position_key = keymapper.add ("NodePosition:");
	forall_nodes (v, this) {
		i++;
		attrs = GAG.node_attrs (v);
		if (attrs != NULL) {
			attr = attrs->find (position_key);
			if ((attr != NULL) && (attr->size () >= 2)) {
				x = attr->head()->get_int ();
				y = attr->tail()->get_int ();
				(*this)[v].p = point (x, y);
				maxx = max (maxx, x);
				maxy = max (maxy, y);
				minx = min (minx, x);
				miny = min (miny, y);
			}
		}
	}
	forall_nodes (v, this) {
		x = ((*this)[v].p.xcoord () - minx) * (90.0 / (maxx - minx)) + 5;
		y = (maxy - (*this)[v].p.ycoord ()) * (90.0 / (maxy - miny)) + 5;
		(*this)[v].p = point (x, y);
	}
	return 0;
}

/*****************************************************************************/

int labelled_graph::write_GraphEd (ostream stream) {
	add_attributes ();
	stream << GAG;
}

/*****************************************************************************/

void labelled_graph::build_labels () {
	//label.init (*this);

	int i = 0;
	node v;
	GAList *attrs;
	GA *attr;
	KeyType label_key = keymapper.add ("LABEL:");
	forall_nodes (v, this) {
		i++;
		attrs = GAG.node_attrs (v);
		if (attrs != NULL) {
			attr = attrs->find (label_key);
		}
		if ((attrs == NULL) || (attr == NULL)) {
			(*this)[v].label = string ("V%d", i);
		} else {
			(*this)[v].label = attr->head()->get_string ();
			if ((*this)[v].label == "") {
				(*this)[v].label = string ("V%d", i);
			}
		}
	}
}

/*****************************************************************************/

void labelled_graph::add_attributes () {
	node v;

	forall_nodes (v, this) {
		// Get the attributes of node v, or
		// create new ones if neccessary
		GAList *attrs = GAG.node_attrs (v);
		if (attrs == nil) {
			GAG.node_attrs (v) = attrs = new GAList ();
		}

		// Add a GA "NodePosition:" with two values
		//GA *attr = new GA (keymapper.add ("NodePosition:"));
		GA *attr = new GA (keymapper.add ("NP"));
		attr->append (new GAValue ((rand () % 400) + 100));
		attr->append (new GAValue ((rand () % 400) + 100));
		attrs->append (attr);
	}
}

/*****************************************************************************/
/* Graph generators                                                          */
/*****************************************************************************/

void petersen_graph (graph &G) {
	node V[10];
	int i;

	G.clear ();
	G.make_undirected ();
	for (i = 0; i < 10; i++) {
		V[i] = G.new_node ();
	}
	for (i = 0; i < 5; i++) {
		G.new_edge (V[i], V[i+5]);
		G.new_edge (V[i], V[(i+1)%5]);
		G.new_edge (V[i+5], V[((i+2) % 5) + 5]);
	}
}

void circulant (graph &G, int n, int c) {
	node *V = new node[n];
	int i, j;

	G.clear ();
	for (i = 0; i < n; i++) {
		V[i] = G.new_node ();
	}
	for (i = 0; i < n; i++) {
		for (j = 1; j <= c; j++) {
			G.new_edge (V[i], V[(i+j) % n]);
		}
	}
}

void cycle (graph &G, int n) {
	circulant (G, n, 1);
}

/*****************************************************************************/
/* Misc                                                                      */
/*****************************************************************************/

bool find_edge (graph &G, edge &e, const node &u, const node &v) {
	forall_adj_edges (e, u) {
		if (G.opposite (u, e) == v) {
			return true;
		}
	}
	return false;
}

bool is_adjacent (graph &G, const node &u, const node &v) {
	edge e;
	return find_edge (G, e, u, v);
}

bool is_clique (graph &G, const list<node> &S) {
	list_item u, v;

	if (S.size () <= 1) {
		return true;
	}

	for (v = u = S.first (); u != S.last (); v = u = S.succ (u)) {
		do {
			v = S.succ (v);
			if (!is_adjacent (G, S.inf (u), S.inf (v))) {
				return false;
			}
		} while (v != S.last ());
	}
	return true;
}

bool is_simplicial (graph &G, node &v) {
	return is_clique (G, G.adj_nodes (v));
}

/*****************************************************************************/
/* Recognition algorithms                                                    */
/*****************************************************************************/

bool is_ktree (graph &G, int k) {
	graph H = G;
	H.make_undirected ();
	node v;
	bool found_simplicial;

	while (H.number_of_nodes () > k) {
		found_simplicial = false;
		forall_nodes (v, H) {
			if (H.degree (v) == k) {
				if (!is_simplicial (H, v)) {
					return false;
				}
				H.del_node (v);
				found_simplicial = true;
				break;
			}
		}
		if (!found_simplicial) {
			return false;
		}
	}
	return is_clique (H, H.all_nodes ());
}

/*****************************************************************************/
/* Graph operators                                                           */
/*****************************************************************************/

#if 0
void complement (graph &newG, graph &G) {
	int n = G.number_of_nodes ();
	int i;
	node_array<int> V (G);

	if (G.is_directed ()) {
		newG.make_directed ();
	} else {
		newG.make_undirected ();
	}

	complete_graph (newG, n);

	node *newV = new node[v];

	i = 0;
	forall_nodes (v, G) {
		V[v] = i;
		newV[i] = v;
		i++;
	}

	




}
#endif

/*****************************************************************************/
/* Graph searches                                                            */
/*****************************************************************************/

/*****************************************************************************/
/* Maximum Cardinality Search                                                */
/* Run time: O(nlogn+m) (could be O(n+m) with bounded priority queues)       */
/*****************************************************************************/

void MCS (graph &G, node &root, node_array<int> &order) {
	node_array<pq_item> q_entry;
	p_queue<int,node> cardinality;
	node v;
	int next_order = 1;

	forall_nodes (v, G) {
		order[v] = 0;
		if (v != root) {
			q_entry[v] = cardinality.insert (0, v);
		}
	}

	for (;;) {
		order[root] = next_order++;
		if (cardinality.empty ()) {
			return;
		}
		forall_adj_nodes (v, root) {
			if (order[v] == 0) {
				cardinality.decrease_p (q_entry[v],
					cardinality.prio (q_entry[v]) - 1);
			}
		}
		root = cardinality.inf (cardinality.find_min ());
		cardinality.del_min ();
	}
}
